/*--

THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
PARTICULAR PURPOSE.

Copyright (C) 1999  Microsoft Corporation.  All rights reserved.

Module Name:

    changepw.c

Abstract:

    Change user's password using the
    Kerberos Change Password Protocol (I-D)

--*/

/*
 * lib/krb5/os/changepw.c
 *
 * Copyright 1990 by the Massachusetts Institute of Technology.
 * All Rights Reserved.
 *
 * Export of this software from the United States of America may
 *   require a specific license from the United States Government.
 *   It is the responsibility of any person or organization contemplating
 *   export to obtain such a license before exporting.
 * 
 * WITHIN THAT CONSTRAINT, permission to use, copy, modify, and
 * distribute this software and its documentation for any purpose and
 * without fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that
 * the name of M.I.T. not be used in advertising or publicity pertaining
 * to distribution of the software without specific, written prior
 * permission.  M.I.T. makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express
 * or implied warranty.
 * 
 */

#define NEED_SOCKETS
#include "k5-int.h"
#include "adm_err.h"
#include "krb5_err.h"
#include "auth_con.h"
#include "kpasswd.h"

#include <stdio.h>
#include <errno.h>

/* Win32 defines. */
#if defined(_WIN32) && !defined(__CYGWIN32__)
#ifndef ECONNABORTED
#define ECONNABORTED WSAECONNABORTED
#endif
#ifndef ECONNREFUSED
#define ECONNREFUSED WSAECONNREFUSED
#endif
#ifndef EHOSTUNREACH
#define EHOSTUNREACH WSAEHOSTUNREACH
#endif
#endif /* _WIN32 && !__CYGWIN32__ */

static const char rcsid[] = "$Id$";

krb5_error_code KRB5_CALLCONV
krb5_mk_chpw_req(context, auth_context, ap_req, passwd, packet)
     krb5_context context;
     krb5_auth_context auth_context;
     krb5_data *ap_req;
     char *passwd;
     krb5_data *packet;
{
    krb5_error_code ret;
    krb5_data clearpw;
    krb5_data cipherpw;
    krb5_replay_data replay;
    char *ptr;

    if (ret = krb5_auth_con_setflags(context, auth_context,
				     KRB5_AUTH_CONTEXT_DO_SEQUENCE))
	return(ret);

    clearpw.length = strlen(passwd);
    clearpw.data = passwd;

    if (ret = krb5_mk_priv(context, auth_context,
			   &clearpw, &cipherpw, &replay))
    return(ret);

    packet->length = 6 + ap_req->length + cipherpw.length;
    packet->data = (char *) malloc(packet->length);
    ptr = packet->data;

    /* length */

    *ptr++ = (packet->length>>8) & 0xff;
    *ptr++ = packet->length & 0xff;

    /* version == 0x0001 big-endian */

    *ptr++ = 0;
    *ptr++ = 1;

    /* ap_req length, big-endian */

    *ptr++ = (ap_req->length>>8) & 0xff;
    *ptr++ = ap_req->length & 0xff;

    /* ap-req data */

    memcpy(ptr, ap_req->data, ap_req->length);
    ptr += ap_req->length;

    /* krb-priv of password */

    memcpy(ptr, cipherpw.data, cipherpw.length);

    return(0);
}

krb5_error_code KRB5_CALLCONV
krb5_rd_chpw_rep(context, auth_context, packet, result_code, result_data)
     krb5_context context;
     krb5_auth_context auth_context;
     krb5_data *packet;
     int *result_code;
     krb5_data *result_data;
{
    char *ptr;
    int plen, vno;
    krb5_data ap_rep;
    krb5_ap_rep_enc_part *ap_rep_enc;
    krb5_error_code ret;
    krb5_data cipherresult;
    krb5_data clearresult;
    krb5_error *krberror;
    krb5_replay_data replay;
    krb5_keyblock *tmp;

    *result_code = 0;
    
    if (packet->length < 4)
	/* either this, or the server is printing bad messages,
	   or the caller passed in garbage */
	return(KRB5KRB_AP_ERR_MODIFIED);

    ptr = packet->data;

    /* verify length */

    plen = (*ptr++ & 0xff);
    plen = (plen<<8) | (*ptr++ & 0xff);

    if (plen != packet->length)
	return(KRB5KRB_AP_ERR_MODIFIED);

    /* verify version number */

    vno = (*ptr++ & 0xff);
    vno = (vno<<8) | (*ptr++ & 0xff);

    if (vno != 1)
	return(KRB5KDC_ERR_BAD_PVNO);

    /* read, check ap-rep length */

    ap_rep.length = (*ptr++ & 0xff);
    ap_rep.length = (ap_rep.length<<8) | (*ptr++ & 0xff);

    if (ptr + ap_rep.length >= packet->data + packet->length)
	return(KRB5KRB_AP_ERR_MODIFIED);

    if (ap_rep.length) {
	/* verify ap_rep */
	ap_rep.data = ptr;
	ptr += ap_rep.length;

	if (ret = krb5_rd_rep(context, auth_context, &ap_rep, &ap_rep_enc))
	    return(ret);

	krb5_free_ap_rep_enc_part(context, ap_rep_enc);

	/* extract and decrypt the result */

	cipherresult.data = ptr;
	cipherresult.length = (packet->data + packet->length) - ptr;

	/* XXX there's no api to do this right. The problem is that
	   if there's a remote subkey, it will be used.  This is
	   not what the spec requires */

	tmp = auth_context->remote_subkey;
	auth_context->remote_subkey = NULL;

	ret = krb5_rd_priv(context, auth_context, &cipherresult, &clearresult,
			   &replay);

	auth_context->remote_subkey = tmp;

	if (ret)
	    return(ret);
    } else {
	cipherresult.data = ptr;
	cipherresult.length = (packet->data + packet->length) - ptr;

	if (ret = krb5_rd_error(context, &cipherresult, &krberror))
	    return(ret);

	clearresult = krberror->e_data;
    }

    if (clearresult.length < 2) {
	ret = KRB5KRB_AP_ERR_MODIFIED;
	goto cleanup;
    }

    ptr = clearresult.data;

    *result_code = (*ptr++ & 0xff);
    *result_code = (*result_code<<8) | (*ptr++ & 0xff);

    if ((*result_code < KRB5_KPASSWD_SUCCESS) ||
	(*result_code > KRB5_KPASSWD_SOFTERROR)) {
	ret = KRB5KRB_AP_ERR_MODIFIED;
	goto cleanup;
    }

    /* all success replies should be authenticated/encrypted */

    if ((ap_rep.length == 0) && (*result_code == KRB5_KPASSWD_SUCCESS)) {
	ret = KRB5KRB_AP_ERR_MODIFIED;
	goto cleanup;
    }

    result_data->length = (clearresult.data + clearresult.length) - ptr;

    if (result_data->length) {
	result_data->data = (char *) malloc(result_data->length);
	memcpy(result_data->data, ptr, result_data->length);
    } else {
	result_data->data = NULL;
    }

    ret = 0;

cleanup:
    if (ap_rep.length) {
	krb5_xfree(clearresult.data);
    } else {
	krb5_free_error(context, krberror);
    }

    return(ret);
}

krb5_error_code KRB5_CALLCONV
krb5_chpw_result_code_string(context, result_code, code_string)
     krb5_context context;
     int result_code;
     char **code_string;
{
   switch (result_code) {
   case KRB5_KPASSWD_MALFORMED:
      *code_string = "Malformed request error";
      break;
   case KRB5_KPASSWD_HARDERROR:
      *code_string = "Server error";
      break;
   case KRB5_KPASSWD_AUTHERROR:
      *code_string = "Authentication error";
      break;
   case KRB5_KPASSWD_SOFTERROR:
      *code_string = "Password change rejected";
      break;
   case KRB5_KPASSWD_ACCESSDENIED:
      *code_string = "Access denied";
      break;
   default:
      *code_string = "Password change failed";
      break;
   }

   return(0);
}

krb5_error_code KRB5_CALLCONV
krb5_change_password(context, creds, newpw, result_code,
		     result_code_string, result_string)
    krb5_context context;
    krb5_creds *creds;
    char *newpw;
    int *result_code;
    krb5_data *result_code_string;
    krb5_data *result_string;
{
    krb5_auth_context auth_context;
    krb5_data ap_req, chpw_req, chpw_rep;
    krb5_address local_kaddr, remote_kaddr;
    krb5_data realm;
    char *host, *code_string;
    krb5_error_code code;
    int i, out, addrlen;
    struct sockaddr *addr_p, local_addr, remote_addr, tmp_addr;
    int cc, local_result_code, tmp_len;
    SOCKET s1, s2;

    auth_context = NULL;

    if (code = krb5_mk_req_extended(context, &auth_context, AP_OPTS_USE_SUBKEY,
				    NULL, creds, &ap_req))
	return(code);

    if ((host = malloc(krb5_princ_realm(context, creds->client)->length + 1))
	== NULL) 
	return ENOMEM;

    strncpy(host, krb5_princ_realm(context, creds->client)->data,
	    krb5_princ_realm(context, creds->client)->length);

    realm.data = host;
    realm.length = strlen(host);

    code = krb5_locate_kpasswd(context, &realm, &addr_p, &out);
    if (code)
	return code;

    if (out == 0) {     /* Couldn't resolve any KDC names */
        krb5_xfree(addr_p);
        return(KADM_NO_HOST);
    }

    /* this is really obscure.  s1 is used for all communications.  it
       is left unconnected in case the server is multihomed and routes
       are asymmetric.  s2 is connected to resolve routes and get
       addresses.  this is the *only* way to get proper addresses for
       multihomed hosts if routing is asymmetric.  

       A related problem in the server, but not the client, is that
       many os's have no way to disconnect a connected udp socket, so
       the s2 socket needs to be closed and recreated for each
       request.  The s1 socket must not be closed, or else queued
       requests will be lost.

       A "naive" client implementation (one socket, no connect,
       hostname resolution to get the local ip addr) will work and
       interoperate if the client is single-homed. */

    if ((s1 = socket(AF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET) {
	krb5_xfree(addr_p);
	return(errno);
    }

    if ((s2 = socket(AF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET) {
	krb5_xfree(addr_p);
	return(errno);
    }

    for (i=0; i<out; i++) {
	if (connect(s2, &addr_p[i], sizeof(addr_p[i])) == SOCKET_ERROR) {
#ifndef HAVE_MACSOCK_H
	    if ((cc < 0) && ((errno == ECONNREFUSED) ||
			     (errno == EHOSTUNREACH)))
#endif
		continue; /* try the next addr */
	    krb5_xfree(addr_p);
	    closesocket(s1);
	    closesocket(s2);
	    return(errno);
	}

	addrlen = sizeof(local_addr);

	if (getsockname(s2, &local_addr, &addrlen) < 0) {
#ifndef HAVE_MACSOCK_H
	    if ((errno == ECONNREFUSED) ||
		(errno == EHOSTUNREACH))
#endif
		continue; /* try the next addr */
	    krb5_xfree(addr_p);
	    closesocket(s1);
	    closesocket(s2);
	    return(errno);
	}

	/* some brain-dead OS's don't return useful information from
	 * the getsockname call.  Namely, windows and solaris.  */

	if (((struct sockaddr_in *)&local_addr)->sin_addr.s_addr != 0) {
	    local_kaddr.addrtype = ADDRTYPE_INET;
	    local_kaddr.length =
	      sizeof(((struct sockaddr_in *) &local_addr)->sin_addr);
	    local_kaddr.contents = 
	      (char *) &(((struct sockaddr_in *) &local_addr)->sin_addr);
	} else {
	    krb5_address **addrs;

	    krb5_os_localaddr(context, &addrs);
	    local_kaddr.magic = addrs[0]->magic;
	    local_kaddr.addrtype = addrs[0]->addrtype;
	    local_kaddr.length = addrs[0]->length;
	    local_kaddr.contents = malloc(addrs[0]->length);
	    memcpy(local_kaddr.contents, addrs[0]->contents, addrs[0]->length);

	    krb5_free_addresses(context, addrs);
	}

	addrlen = sizeof(remote_addr);
	if (getpeername(s2, &remote_addr, &addrlen) < 0) {
#ifndef HAVE_MACSOCK_H
	    if ((errno == ECONNREFUSED) ||
		(errno == EHOSTUNREACH))
#endif
		continue; /* try the next addr */
	    krb5_xfree(addr_p);
	    closesocket(s1);
	    closesocket(s2);
	    return(errno);
	}

	remote_kaddr.addrtype = ADDRTYPE_INET;
	remote_kaddr.length =
	    sizeof(((struct sockaddr_in *) &remote_addr)->sin_addr);
	remote_kaddr.contents = 
	    (char *) &(((struct sockaddr_in *) &remote_addr)->sin_addr);

	/* mk_priv requires that the local address be set.
	  getsockname is used for this.  rd_priv requires that the
	  remote address be set.  recvfrom is used for this.  If
	  rd_priv is given a local address, and the message has the
	  recipient addr in it, this will be checked.  However, there
	  is simply no way to know ahead of time what address the
	  message will be delivered *to*.  Therefore, it is important
	  that either no recipient address is in the messages when
	  mk_priv is called, or that no local address is passed to
	  rd_priv.  Both is a better idea, and I have done that.  In
	  summary, when mk_priv is called, *only* a local address is
	  specified.  when rd_priv is called, *only* a remote address
	  is specified.  Are we having fun yet?  */

	if (code = krb5_auth_con_setaddrs(context, auth_context, &local_kaddr,
					  NULL)) {
	    krb5_xfree(addr_p);
	    closesocket(s1);
	    closesocket(s2);
	    return(code);
	}

	if (code = krb5_mk_chpw_req(context, auth_context, &ap_req,
				    newpw, &chpw_req)) {
	    krb5_xfree(addr_p);
	    closesocket(s1);
	    closesocket(s2);
	    return(code);
	}

	if ((cc = sendto(s1, chpw_req.data, chpw_req.length, 0,
			 (struct sockaddr *) &addr_p[i],
			 sizeof(addr_p[i]))) !=
	    chpw_req.length) {
#ifndef HAVE_MACSOCK_H
	    if ((cc < 0) && ((errno == ECONNREFUSED) ||
			     (errno == EHOSTUNREACH)))
#endif
		continue; /* try the next addr */
	    krb5_xfree(addr_p);
	    closesocket(s1);
	    closesocket(s2);
	    return((cc < 0)?errno:ECONNABORTED);
	}

	krb5_xfree(chpw_req.data);

	chpw_rep.length = 1500;
	chpw_rep.data = (char *) malloc(chpw_rep.length);

	/* XXX need a timeout/retry loop here */

	/* "recv" would be good enough here... except that Windows/NT
	   commits the atrocity of returning -1 to indicate failure,
	   but leaving errno set to 0.
	   
	   "recvfrom(...,NULL,NULL)" would seem to be a good enough
	   alternative, and it works on NT, but it doesn't work on
	   SunOS 4.1.4 or Irix 5.3.  Thus we must actually accept the
	   value and discard it. */
	tmp_len = sizeof(tmp_addr);
	if ((cc = recvfrom(s1, chpw_rep.data, chpw_rep.length, 0, &tmp_addr, &tmp_len)) < 0) {
	    krb5_xfree(addr_p);
	    closesocket(s1);
	    closesocket(s2);
	    return(errno);
	}

	closesocket(s1);
	closesocket(s2);

	chpw_rep.length = cc;

	if (code = krb5_auth_con_setaddrs(context, auth_context, NULL,
					  &remote_kaddr)) {
	    krb5_xfree(addr_p);
	    closesocket(s1);
	    closesocket(s2);
	    return(code);
	}

	code = krb5_rd_chpw_rep(context, auth_context, &chpw_rep,
				&local_result_code, result_string);

	free(chpw_rep.data);
	krb5_xfree(addr_p);

	if (code && !local_result_code)
	    return(code);

	if (result_code)
	    *result_code = local_result_code;

	if (result_code_string) {
	    if (code = krb5_chpw_result_code_string(context, local_result_code,
						    &code_string))
		return(code);

	    result_code_string->length = strlen(code_string);
	    if ((result_code_string->data =
		 (char *) malloc(result_code_string->length)) == NULL)
		return(ENOMEM);
	    strncpy(result_code_string->data, code_string,
		    result_code_string->length);
	}

	return(0);
    }

    krb5_xfree(addr_p);
    closesocket(s1);
    closesocket(s2);
    return(errno);
}
